@@ -171,7 +171,8 @@ def render(box_size,
            geometry,
            buffer_size=None,
            background_color=None,
-           resolution=None):
+           resolution=None,
+           frame_rate=None):
   """Creates a rendering front-end along with callbacks in the host program.
 
   Args:
@@ -184,6 +185,8 @@ def render(box_size,
     background_color: An array of shape (3,) specifying the background color of
       the visualization.
     resolution: The resolution of the renderer.
+    frame_rate: An optional integer specifying the target frames-per-second
+      for the renderer.
   """
   global SIMULATION_IDX
   # INTERNAL_RENDERER_CODE_LOADING
@@ -194,8 +197,8 @@ def render(box_size,
   dimension = None
 
   if not isinstance(geometry, dict):
-    geometry = { 'all': geometry } 
-  
+    geometry = { 'all': geometry }
+ 
   for geom in geometry.values():
     if hasattr(geom, 'position'):
       assert dimension is None or geom.position.shape[-1] == dimension
@@ -236,6 +239,9 @@ def render(box_size,
     if resolution is not None:
       metadata['resolution'] = resolution
 
+    if frame_rate is not None:
+      metadata['frame_rate'] = frame_rate
+
     return _to_json(metadata)
   output.register_callback('GetSimulationMetadata', get_metadata)
 
